from sympy import Matrix, lcm
import re

def balance_equation(reactants, products):
    
    #completely AI generated and smh works, don`t touch anything here unless you really know what you are doing, cuz i don`t give a fuck what anything does
    element_list = []
    element_matrix = []

    def add_to_matrix(element, index, count, side):
        if index == len(element_matrix):
            element_matrix.append([0]*len(element_list))
        if element not in element_list:
            element_list.append(element)
            for row in element_matrix:
                row.append(0)
        column = element_list.index(element)
        element_matrix[index][column] += count*side

    def find_elements(segment, index, multiplier, side):
        elements_and_numbers = re.split('([A-Z][a-z]?)', segment)
        i = 0
        while i < len(elements_and_numbers) - 1:  # last element always blank
            i += 1
            if len(elements_and_numbers[i]) > 0:
                if elements_and_numbers[i+1].isdigit():
                    count = int(elements_and_numbers[i+1])*multiplier
                    add_to_matrix(elements_and_numbers[i], index, count, side)
                    i += 1
                else:
                    add_to_matrix(elements_and_numbers[i], index, multiplier, side)

    def compound_decipher(compound, index, side):
        segments = re.split('(\([A-Za-z0-9]*\)[0-9]*)', compound)
        for segment in segments:
            if segment.startswith("("):
                segment = re.split('\)([0-9]*)', segment)
                multiplier = int(segment[1])
                segment = segment[0][1:]
            else:
                multiplier = 1
            find_elements(segment, index, multiplier, side)

    for i, reactant in enumerate(reactants):
        compound_decipher(reactant, i, 1)
    for i, product in enumerate(products):
        compound_decipher(product, i+len(reactants), -1)

    # Solve the system
    element_matrix = Matrix(element_matrix).transpose()
    solution = element_matrix.nullspace()[0]

    # Find the least common multiple of the denominators of all the fractions in solution
    multiple = lcm([val.q for val in solution])
    solution = multiple*solution

    # Convert solution to a list of coefficients
    coefficients = solution.tolist()

    # Format the output
    output = " + ".join([f"{coefficients[i][0]}{reactants[i]}" for i in range(len(reactants))])
    output += " => "
    output += " + ".join([f"{coefficients[i+len(reactants)][0]}{products[i]}" for i in range(len(products))])

    return output


import re
from requests import get as gt
from os import listdir
import json
from functools import lru_cache

@lru_cache(maxsize=32)
def get_elements(lower_bound:int = 0,upper_bound: int = 100,raw: bool = False):

    #check if bounds are correct and if not correct them
    if upper_bound<lower_bound:
        temp = upper_bound
        upper_bound = lower_bound
        lower_bound = temp
    if upper_bound>118:
        upper_bound=118
    if lower_bound<0:
        lower_bound=0
        
    
    #download json containig all info and if it`s already downloaded use it

    if not "chem_elements.json" in listdir():
        response = gt("https://raw.githubusercontent.com/Bowserinator/Periodic-Table-JSON/master/PeriodicTableJSON.json")
        #error while accessing
        if response.status_code != 200:
            raise("Error accessing remote json")
        data = response.json()
        #save for later offline uses
        with open("chem_elements.json", 'w') as file:
            json.dump(response.json(), file)
    #so the json exists and just need to be opened
    else:
        data = open("chem_elements.json")
        data = json.load(data)
    #return first [lower_bound-upper_bound] ammount of elemnts, default is 100, can be lowered to inscrease speed
    if raw == False:
        return [element['symbol'] for element in data['elements'] if lower_bound <= element['number'] <= upper_bound]
    else:
        return data
@lru_cache(maxsize=32)
def molar_mass(formula: str) -> float:
    parsed_formula = parse_chemical_formula(formula)
    molar = 0.0
    data = get_elements(raw=True)
    element_to_mass = {element['symbol']: element['atomic_mass'] for element in data['elements']}
    for element, count in parsed_formula.items():
        molar += float(element_to_mass[element]) * count
    return molar
@lru_cache(maxsize=32)
def parse_chemical_formula(formula:str,error_finding: bool = True) -> dict:
    #delete all spaces cuz they don`t have meaning in them and can make function do wrong outputs
    formula = formula.replace(" ","")
    # Step 1: Extract and remove the whole molecule coefficient
    coefficient = 1
    for i, char in enumerate(formula):
        if not char.isdigit():
            coefficient = int(formula[:i] or 1)
            formula = formula[i:]
            break

    # Step 2: Parse the chemical elements.
    element_pattern = re.compile(r"([A-Z][a-z]*)(\d*)|(\()([^)]+)(\))(\d*)")
    # Define dictionary to hold results
    result = {}

    formula_parts = element_pattern.findall(formula)

    for part in formula_parts:
        element, primary_number, _, sub_formula, _, multiplier = part
        if element:
            primary_number = int(primary_number) if primary_number else 1
            primary_number *= coefficient # multiply count by the molecule's coefficient
            result[element] = result.get(element, 0) + primary_number
        else:
            sub_coefficient = coefficient * (int(multiplier) if multiplier else 1)
            # Process the sub-formula recursively
            sub_result = parse_chemical_formula(sub_formula)
            for sub_element, sub_value in sub_result.items():
                sub_value *= sub_coefficient
                result[sub_element] = result.get(sub_element, 0) + sub_value

    if error_finding:
        k = get_elements()
        for i, l in result.items():
            if not i in k:
                raise ValueError(f"Error while parsing formula '{formula}': element '{i}' not found in element list")
            
    return result


@lru_cache(maxsize=32)
def check_equasion(eq:str) -> bool:
    #taking in mind that eq should have this pattern "a + b => c + d" or "a=>b+c+d+e"
    first,second = eq.split("=>")
    def count_half(hf:str) -> dict:
        res = {}
        hf = hf.split("+")
        for l in hf:
            for k,v in parse_chemical_formula(l).items():
                if k in res:
                    res[k] += v
                else:
                    res[k] = v
        return res
    if count_half(first)==count_half(second):
        return True
    return False






from setuptools import setup
setup(
    name='chemtools-py',
    version='0.0.2',    
    description='A example Python package',
    long_description="somewhat in the future",
    url='https://github.com/KerrKerrKerr/chemtools-py',
    download_url="https://github.com/KerrKerrKerr/chemtools-py.git",
    author='burntSynaps3',
    author_email='ustinov2005@example.com',
    license='BRUH',
    packages=['chemtools-py'],
    install_requires=[
        "sympy"
    ],
)
